from collections import OrderedDict

class LabeledTreeCorpus(list):
    """
    Read in the Stanford Sentiment Treebank using the original serialization format:

    > (3 (2 this) (3 (2 is) (3 good ) )

    """

    def labels(self):
        labelings = OrderedDict()
        for tree in self:
            for label, line in tree.to_labeled_lines():
                labelings[line] = label
        return labelings

    def to_file(self, path, mode = "w"):
        with open(path, mode = mode) as f:
            for tree in self:
                for label, line in tree.to_labeled_lines():
                    f.write(line + "\n")



"""
Make trees visualizable in an IPython notebook
"""
import pyximport
pyximport.install()
import json

try:
    from PIL import ImageFont
    from IPython.display import Javascript, display
    font = ImageFont.core.getfont("/Library/Fonts/Georgia.ttf", 15)
    def text_size(text):
        return max(4, font.getsize(text)[0][0])
except ImportError:
    def text_size(text):
        return max(4, int(len(s) * 1.1))
# make changes here to incorporate cap and uncap unknown words.

def dot_product(a, b):
    if type(a) != list:
        a = list(a)
    if type(b) != list:
        b = list(b)
    out = 0.0
    for a_x, b_x in zip(a, b):
        out += a_x * b_x
    return out

class LabeledSentence(object):
    def __init__(self, sentence):
        self.sentence = sentence
        self.label    = None

class LabeledTree(object):
    def __init__(self, depth = 0, sentence = None, label = None, children = None, parent = None, udepth = 1, calculated_embedding = None):
        self.label    = label
        self.predicted_label = None
        self.children = children if children != None else []
        self.general_children = []
        self.sentence = sentence
        self.calculated_embedding = calculated_embedding
        self.parent   = parent
        self.depth    = depth
        self.udepth   = udepth

    def uproot(tree):
        uprooted = tree.copy()
        uprooted.parent = None
        for child in tree.all_children():
            uprooted.add_general_child(child)
        return uprooted

    def shrink_tree(tree, final_depth):
        if tree.udepth <= final_depth:
            return tree
        for branch in tree.general_children:
            if branch.udepth == final_depth:
                return branch.uproot()

    def shrunk_trees(tree, final_depth):
        if tree.udepth <= final_depth:
            yield tree
        for branch in tree.general_children:
            if branch.udepth == final_depth:
                yield branch.uproot()

    def copy(self):
        return LabeledTree(
            calculated_embedding = self.calculated_embedding,
            udepth = self.udepth,
            depth = self.depth,
            sentence = self.sentence,
            label = self.label,
            children = self.children.copy() if self.children != None else [],
            parent = self.parent)

    def add_child(self, child):
        self.children.append(child)
        # assert(child.depth == self.depth + 1)
        child.parent = self

    def add_general_child(self, child):
        self.general_children.append(child)

    def all_children(self):
        if len(self.children) > 0:
            for child in self.children:
                for subchild in child.all_children():
                    yield subchild
            yield self
        else:
            yield self

    def lowercase(self):
        if len(self.children) > 0:
            for child in self.children:
                child.lowercase()
        else:
            self.sentence = self.sentence.lower()

    def true_label(self, model):
        return model.true_label(self)

    def to_dict(self, model, index = 1, prediction = True):
        """
        Dict format for use in Javascript / Jason Chuang's display technology.
        """
        rep = {}
        rep["index"] = index
        rep["leaf"] = len(self.children) == 0
        rep["depth"] = self.udepth
        if prediction:
            pred_label = self.predicted_label
        else:
            pred_label = self.true_label(model)
        rep["scoreDistr"] = pred_label.tolist()
        rep["rating"] = dot_product(pred_label, [-12.5,-6.25,0.0,6.25,12.5])+12.5
        rep["numChildren"] = len(self.children)
        text = self.sentence if self.sentence != None else ""
        seen_tokens = 0
        witnessed_pixels = 0
        for i, child in enumerate(self.children):
            if i > 0:
                text += " "
            child_key = "child%d" % (i)
            index += 1
            rep[child_key] = child.to_dict(model, index = index, prediction = prediction)
            text += rep[child_key]["text"]
            seen_tokens += rep[child_key]["tokens"]
            witnessed_pixels += rep[child_key]["pixels"]

        rep["text"] = text
        rep["tokens"] = 1 if (self.sentence != None and len(self.sentence) > 0) else seen_tokens
        rep["pixels"] = witnessed_pixels + 3 if len(self.children) > 0 else text_size(self.sentence)
        return rep

    def to_json(self, model, prediction = True):
        return json.dumps(self.to_dict(model, prediction = prediction))

    def display(tree, model, prediction = True):
        if prediction:
            model.forward_propagate(tree)
        display(Javascript("createTrees(["+tree.to_json(model, prediction = prediction)+"])"))
        display(Javascript("updateTrees()"))

    def to_lines(self):
        if len(self.children) > 0:
            left_lines, right_lines = self.children[0].to_lines(), self.children[1].to_lines()
            self_line = [left_lines[0] + " " + right_lines[0]]
            return self_line + left_lines + right_lines
        else:
            return [self.sentence]

    def to_labeled_lines(self):
        if len(self.children) > 0:
            left_lines, right_lines = self.children[0].to_labeled_lines(), self.children[1].to_labeled_lines()
            self_line = [(self.label, left_lines[0][1] + " " + right_lines[0][1])]
            return self_line + left_lines + right_lines
        else:
            return [(self.label, self.sentence)]


    def total_projections_predictions(self, model = None):
        return total_projections_predictions(self, model)

    def __str__(self):
        """
        String representation of a tree as visible in original corpus.

        print(tree)
        #=> '(2 (2 not) (3 good))'

        Outputs
        -------

            str: the String representation of the tree.

        """
        if len(self.children) > 0:
            rep = "(%d " % self.label
            for child in self.children:
                rep += str(child)
            return rep + ")"
        else:
            return ("(%d %s) " % (self.label, self.sentence))



import re
from .labeled_trees       import LabeledTree
from .labeled_tree_corpus import LabeledTreeCorpus

find_bubbles = re.compile(r'(\([^\(\)]+\))')

class ParseError(Exception):
    def __init__(self, message):
        self.message = message

def create_leaves_from_string(line):
    matches = re.findall(find_bubbles, line)
    for match in matches:
        yield create_tree_from_string(match)

def create_tree_from_string(line):
    depth         = 0
    awaiting_num  = False
    current_word  = ""
    root          = None
    current_node  = root

    for char in line:

        if awaiting_num:

            current_node.label = int(char)
            awaiting_num = False

        else:
            if char == '(':
                depth += 1
                if depth > 1:
                    # replace current head node by this node:
                    child = LabeledTree(depth = depth)
                    current_node.add_child(child)
                    current_node = child
                    root.add_general_child(child)
                else:
                    root = LabeledTree(depth = depth)
                    root.add_general_child(root)
                    current_node = root

                awaiting_num = True

            elif char == ')':

                # assign current word:
                if len(current_word) > 0:
                    current_node.sentence = current_word.replace("\xa0", " ")
                    current_node.udepth   = 1

                    # erase current word
                    current_word = ""

                # go up a level:
                depth -= 1
                if current_node.parent != None:
                    current_node.parent.udepth = max(current_node.udepth+1, current_node.parent.udepth)
                current_node = current_node.parent

            elif char == " ":
                # ignore spacing
                continue
            else:
                # add to current read word
                current_word += char
    if depth != 0:
        raise ParseError("Not an equal amount of closing and opening parentheses")

    return root

def check_udepth(tree):
    depths = depth_sorted_children(tree)
    for i in reversed(range(1, depths[-1]+1)):
        for tree in depths[i]:
            if len(tree.children) == 0:
                assert(tree.udepth == 1)
            else:
                assert(tree.udepth == max([i.udepth for i in tree.children]) + 1)
    return tree

def import_tree_corpus_words(trees):
    tree_list = LabeledTreeCorpus()
    with open(trees, "rt") as f:
        for line in f:
            for tree in create_leaves_from_string(line):
                tree_list.append(tree)
    return tree_list

def import_tree_corpus(trees):
    tree_list = LabeledTreeCorpus()
    with open(trees, "rt") as f:
        for line in f:
            tree_list.append(create_tree_from_string(line))
    return tree_list
